// Testing code: implementation of DerivedEcho objects
//
//   interface Echo {
//        string echoString(in string mesg);
//   };
//
//   interface DerivedEcho : Echo {
//        long echoLong (in long arg1);
//   };

#include <iostream.h>
#include <string.h>
#include <derivedEcho.hh>
#include <common/common.h>

class Echo_i : public virtual _sk_Echo {
public:
  Echo_i() {}
  virtual char * echoString(const char *mesg);

};

char *
Echo_i::echoString(const char *mesg) {
  char *p = CORBA::string_alloc(strlen((const char*)mesg));
  strcpy((char *)p,(const char *)mesg);
  return p;
}

// implementation using inheritance skeleton
class DerivedEcho_i : public virtual Echo_i, public virtual _sk_DerivedEcho {
public:
  DerivedEcho_i() {}
  //  virtual char * echoString(const char *mesg);
  virtual CORBA::Long echoLong(CORBA::Long arg1);

};

#if 0
char *
DerivedEcho_i::echoString(const char *mesg) {
  char *p = CORBA::string_alloc(strlen((const char*)mesg));
  strcpy((char *)p,(const char *)mesg);
  return p;
}

#endif

CORBA::Long
DerivedEcho_i::echoLong(CORBA::Long arg1) {
  return arg1;
}

int
main(int argc, char **argv) {

  try {

    CORBA::ORB_ptr orb = CORBA::ORB_init(argc,argv,"omniORB2");
    CORBA::BOA_ptr boa = orb->BOA_init(argc,argv,"omniORB2_BOA");

    CosNaming::Name use_name;

    if (argc > 1) {
      if (argc != 2) {
	cerr << "Usage: derivedEchosrv [<COSNaming compound name (e.g. x/y/echo.obj)>]"	
	     << endl;
      }
      try {
	use_name = string_to_name(argv[1]);
      }
      catch (...) {
	cerr << argv[1] << " is not a valid COS Naming compound name." << endl;
	return 1;
      }
    }

    DerivedEcho_i *myobj = new DerivedEcho_i();
    myobj->_obj_is_ready(boa);

    {
      DerivedEcho_var myobjRef = myobj->_this();
      if (use_name.length()) {
	if (!bindObjectToName(orb,myobjRef,use_name)) {
	  return 1;
	}
      }
      else {
	CORBA::String_var p = orb->object_to_string(myobjRef);
	cerr << (char*)p << endl;
      }
    }

    boa->impl_is_ready();

  }
  catch (const omniORB::fatalException &ex) {
      cerr << "#### You have caught an omniORB2 bug, details are as follows:" << endl;
      cerr << ex.file() << " " << ex.line() << ":" << ex.errmsg() << endl; 
      cerr << "main thread exited." << endl;
      return 1;
    }
  return 0;
}
