// Testing code: implementation of echo objects
//
//   interface Echo {
//        string echoString(in string mesg);
//   };

#include <iostream.h>
#include <string.h>
#include <echo.hh>
#include <common/common.h>

// implementation using inheritance skeleton
class Echo_i : public virtual _sk_Echo {
public:
  Echo_i() {}
  virtual char * echoString(const char *mesg);

};

char *
Echo_i::echoString(const char *mesg) {
  char *p = CORBA::string_alloc(strlen((const char*)mesg));
  strcpy((char *)p,(const char *)mesg);
  return p;
}

int
main(int argc, char **argv)
{
  CORBA::ORB_var orb = CORBA::ORB_init(argc,argv,"omniORB2");
  CORBA::BOA_var boa = orb->BOA_init(argc,argv,"omniORB2_BOA");

  CosNaming::Name use_name;
  
  if (argc > 1) {
    if (argc != 2) {
      cerr << "Usage: echosrv [<COSNaming compound name (e.g. x/y/echo.obj)>]"	
	   << endl;
    }
    try {
      use_name = string_to_name(argv[1]);
    }
    catch (...) {
      cerr << argv[1] << " is not a valid COS Naming compound name." << endl;
      return 1;
    }
  }

  Echo_i *myobj = new Echo_i();
  myobj->_obj_is_ready(boa);

  {
    Echo_var myobjRef = myobj->_this();
    if (use_name.length()) {

      try {
	if (!bindObjectToName(orb,myobjRef,use_name)) {
	  return 1;
	}
      }
      catch (...) {
	cerr << "bindObjectToName returns a system exception." << endl;
	return 1;
      }
    }
    else {
      CORBA::String_var p = orb->object_to_string(myobjRef);
      cerr << (char*)p << endl;
    }
  }

  boa->impl_is_ready();
  // Tell the BOA we are ready. The BOA's default behaviour is to block
  // on this call indefinitely.

  return 0;
}
